#include "data.h"
#include "error.h"
#include "hashtbl.h"
#include "globals.h"

coord x0 = 0, y0 = 0;
int unrestricted = 0;

#define Absolutize(x, y) {x = (x + x0) % (1 << 27); y = (y + y0) % (1 << 27);}

int com_read (result, x, y)
  brick_t result;
  coord x, y;
{
  bucket_t b;
  block_list_t bl;

  Absolutize (x, y);
  if (!lock_bucket (b, hash (x, y))){
    error = E_cant_lock;
    return;
  }
  find_bl (bl, b, x, y);
  unpack_brick (result, bl);
  ++t_read;
  if (bl->length == 0) ++t_empty;
  unlock_bucket (b);
  return (bl->length == 0);
}

/* There must be no locked bucket when calling this function. */
static void free_random_brick ()
{
  bucket_addr a;
  bucket_t b;
  block_list_t bl;
  coord x, y;
  int d;

  /* Iterate until success. */
  while (1){
    a = random (I_num_buckets);
    if (!lock_bucket (b, a)) continue;
    if (find_last (&x, &y, b)){
      remove_bl (bl, &d, b, x, y);
      if (x < I_ro_x && y < I_ro_y){
	insert_bl (b, bl, 0);
	unlock_bucket (b);
	continue;
      }else{
	free_blocks (bl, bl->length);
	unlock_bucket (b);
	return;
      }
    }
  }
}

void com_write (x, y, brick)
  coord x, y;
  brick_t brick;
{
  bucket_t b;
  block_list_t bl;
  long diff;
  brick_t old_brick;
  int holes = has_holes (brick);
  int head_of_bucket;
  bucket_addr h;
  int depth;

  Absolutize (x, y);
  if (x < I_ro_x && y < I_ro_y && !unrestricted){
    error = E_read_only_brick;
    return;
  }
  h = hash (x, y);
  while (1){
    if (!lock_bucket (b, h)){
      error = E_cant_lock;
      return;
    }
    remove_bl (bl, &depth, b, x, y);
    /* Never put a brick in front of the central brick. */
    if (h == 0) depth = 1;
    /* Put the central brick in front of all others. */
    if (x < I_ro_x && y < I_ro_y) depth = 0;
    if (bl->length == 0) ++t_added;
    if (holes){
      unpack_brick (old_brick, bl);
      merge_bricks (brick, old_brick);
    }
    if (is_empty (brick)){
      free_blocks (bl, bl->length);
      ++t_erased;
      break;
    }
    diff = pack_brick (bl, brick);
    if (diff < 0){
      free_blocks (bl, -diff);
    }else if (diff > 0){
      if (!alloc_blocks (bl, diff)){
        unlock_bucket (b);
        free_random_brick ();
        ++t_blasted;
        continue;
      }
    }
    insert_bl (b, bl, (depth <= 0) ? 0 : 1);
    break;
  }
  ++t_written;
  unlock_bucket (b);
}

void com_warp ()
{
  block_list_t bl;
  unsigned long num;
  bucket_t b;

  num = random (I_num_buckets);
  while (1){
    if (lock_bucket (b, num)){
      if (find_first (&x0, &y0, b)) break;
      unlock_bucket (b);
    }
    num = (num + 1) % I_num_buckets;
  }
  unlock_bucket (b);
}
