#include <sys/types.h>
#include <unistd.h>
char *malloc ();

#include "data.h"
#include "globals.h"

void read_block (result, addr)
     block_t result;
     block_addr addr;
{
  io_read (data_fd, addr, sizeof (struct disk_block), (char *) result);
  result->self = addr;
}

void write_block (b)
     block_t b;
{
  io_write (data_fd, b->self, sizeof (struct disk_block), (char *) b);
}

int set_num_blocks (num_blocks)
     long num_blocks;
{
  long new_size = num_blocks * sizeof (struct disk_block);
  long old_size = lseek (data_fd, 0L, SEEK_END);

  if (new_size < old_size){
    ftruncate (data_fd, new_size);
  }else{
    char *buf = malloc (new_size - old_size);
    if (buf == NULL) return 0;
    if (write (data_fd, buf, new_size - old_size) != new_size - old_size){
      ftruncate (data_fd, old_size);
      return 0;
    }
  }
  return 1;
}

void read_desc (result, addr)
  block_list_t result;
  block_addr addr;
{
  io_read (data_fd, addr, sizeof (struct disk_block), (char *) Desc (result));
  Desc (result)->self = addr;
}

void read_blocks (bl)
     block_list_t bl;
{
  block_addr a;
  long i;

  a = Desc (bl)->next;
  for (i = 1; i < 18; i++){
    if (a == Nil){
      bl->length = i;
      return;
    }
    read_block (&(bl->blocks [i]), a);
    a = bl->blocks [i].next;
  }
  bl->length = 18;
  if (a != Nil) log ("Too many blocks in a brick, block = ", a);
}

void check_blocks (bl, bitmap)
     block_list_t bl;
     char *bitmap;
{
  block_addr a;
  long i;

  a = Desc (bl)->next;
  for (i = 1; i < 18; i++){
    if (a == Nil){
      bl->length = i;
      return;
    }
    bitmap [(bl->blocks [i]).self] |= 4;
    read_block (&(bl->blocks [i]), a);
    a = bl->blocks [i].next;
  }
  if (a != Nil) printf ("long brick, extra block %x\n", a);
}

int is_empty (brick)
  brick_t brick;
{
  long i, j;

  for (i = 0; i < 32; i++){
    for (j = 0; j < 32; j++){
      if (brick [i] [j] != ' ') return 0;
    }
  }
  return 1;
}

void fill_brick (result)
  brick_t result;
{
  long i, j;

  for (i = 0; i < 32; i++){
    for (j = 0; j < 32; j++){
      result [i] [j] = ' ';
    }
  }
}

int has_holes (brick)
  brick_t brick;
{
  long i, j;

  for (i = 0; i < 32; i++){
    for (j = 0; j < 32; j++){
      if (brick [i] [j] == '\177') return 1;
    }
  }
  return 0;
}

void merge_bricks (brick, old_brick)
     brick_t brick, old_brick;
{
  long i, j;

  for (i = 0; i < 32; i++){
    for (j = 0; j < 32; j++){
      if (brick [i] [j] == '\177') brick [i] [j] = old_brick [i] [j];
    }
  }
}
