#ifndef _data_
#define _data_

typedef unsigned long coord;
typedef long block_addr;
typedef unsigned long bucket_addr;

/* A bucket on disk. */
struct disk_bucket {
  block_addr first;
};

/* A bucket in memory. */
typedef struct bucket {
  block_addr first;
  bucket_addr self;
} bucket_t [1];

/* A block on disk. */
struct disk_block {
  block_addr next;
  char data [60];
};

/* A block with descriptor in memory. */
typedef struct desc_block {
  block_addr next;
  coord x, y;
  block_addr next_desc;
  char data [48];
  block_addr self;
} desc_block_t [1];

/* A normal block in memory. */
typedef struct block {
  block_addr next;
  char data [60];
  block_addr self;
} block_t [1];

/* A brick (unpacked). */
typedef char brick_t [32] [32];
#define Brick_size sizeof (brick_t)

#define Blocks_per_brick 18

/* A brick (in block list form). */
typedef struct block_list {
  long length;
  struct block blocks [Blocks_per_brick];
} block_list_t [1];
#define Desc(bl) ((struct desc_block *) &(bl->blocks [0]))

#define Nil ((block_addr) -1)

void read_block ();
void write_block ();
int set_num_blocks ();
void read_desc ();
void read_blocks ();
void check_blocks ();
int is_empty ();
void fill_brick ();
int has_holes ();
void merge_bricks ();

#endif /* _data_ */
