#include "data.h"
#include "io.h"
#include "globals.h"

static void grab_free_list ()
{
  if (io_lock (misc_fd, O_free_head, sizeof (I_free_head))){
    io_read (misc_fd, O_free_head, sizeof (I_free_head),
	     (char *) &I_free_head);
    io_read (misc_fd, O_free_length, sizeof (I_free_length),
	     (char *) &I_free_length);
    io_read (misc_fd, O_cur_blocks, sizeof (I_cur_blocks),
	     (char *) &I_cur_blocks);
  }else{
    bug ("grab_free_list");
  }
}

static void release_free_list ()
{
  block_t b;
  int trunc_flag = 0;

  while (I_free_head == I_cur_blocks - 1){
    read_block (b, I_free_head);
    I_free_head = b->next;
    --I_free_length;
    --I_cur_blocks;
    trunc_flag = 1;
  }
  io_write (misc_fd, O_cur_blocks, sizeof (I_cur_blocks),
	    (char *) &I_cur_blocks);
  if (trunc_flag) set_num_blocks (I_cur_blocks);
  io_write (misc_fd, O_free_head, sizeof (I_free_head), (char *) &I_free_head);
  io_write (misc_fd, O_free_length, sizeof (I_free_length),
	    (char *) &I_free_length);
  io_unlock (misc_fd, O_free_head, sizeof (I_free_head));
}

int new_blocks (n)
     long n;
{
  long i;
  block_t b;

  if (!set_num_blocks (I_cur_blocks + n)) return 0;
  for (i = I_cur_blocks; i < I_cur_blocks + n; i++){
    b->self = i;
    b->next = i-1;
    write_block (b);
  }
  b->self = I_cur_blocks;
  b->next = I_free_head;
  write_block (b);
  I_free_length += n;
  I_cur_blocks += n;
  I_free_head = I_cur_blocks - 1;
  return 1;
}

/* Free the last [n] blocks. */
void free_blocks (bl, n)
     block_list_t bl;
     long n;
{
  if (n == 0) return;
  grab_free_list ();
  if (bl->length - n > 0){
    bl->blocks [bl->length - n - 1].next = Nil;
    write_block (&(bl->blocks [bl->length - n - 1]));
  }
  bl->blocks [bl->length - 1].next = I_free_head;
  write_block (&(bl->blocks [bl->length - 1]));
  I_free_head = bl->blocks [bl->length - n].self;
  I_free_length += n;
  release_free_list ();
  bl->length -= n;
}

int alloc_blocks (bl, n)
  block_list_t bl;
  long n;
{
  int retcode;
  block_t b;
  long i;

  grab_free_list ();
  if (I_max_blocks + I_free_length < n + I_cur_blocks){
    release_free_list ();
    return 0;
  }else{
    if (I_free_length < n){
      if (!new_blocks (n - I_free_length)){
	release_free_list ();
	return 0;
      }
    }
    for (i = bl->length; i < bl->length + n; i++){
      read_block (b, I_free_head);
      I_free_head = b->next;
      bl->blocks [i].next = b->next;
      bl->blocks [i].self = b->self;
    }
    I_free_length -= n;
    release_free_list ();
    bl->blocks [bl->length + n - 1].next = Nil;
    if (bl->length > 0){
      bl->blocks [bl->length - 1].next = bl->blocks [bl->length].self;
    }
    bl->length += n;
    return 1;
  }
}

void verify_free_list (bitmap, verbose)
     char *bitmap;
     int verbose;
{
  block_addr a;
  block_t b;
  long n = 0;

  grab_free_list ();
  a = I_free_head;
  while (a != Nil){
    ++n;
    read_block (b, a);
    bitmap [b->self] |= 1;
    if (verbose) printf ("%x.", a);
    a = b->next;
  }
  if (n < I_free_length){
    printf ("The free list is too short. (%d)\n", I_free_length - n);
  }else if (n > I_free_length){
    printf ("The free list is too long. (%d)\n", n - I_free_length);
  }
  release_free_list ();
}
