#include "config.h"
#include "data.h"
#include "error.h"
#include "io.h"
#include "globals.h"

int lock_bucket (result, addr)
  bucket_t result;
  bucket_addr addr;
{
  if (!io_lock (table_fd, addr, sizeof (struct disk_bucket))){
    return 0;
  }
  io_read (table_fd, addr, sizeof (struct disk_bucket), (char*) result);
  result->self = addr;
  return 1;
}

static void write_bucket (b)
  bucket_t b;
{
  io_write (table_fd, b->self, sizeof (struct disk_bucket), (char *) b);
}

void unlock_bucket (b)
  bucket_t b;
{
  io_unlock (table_fd, b->self, sizeof (struct disk_bucket));
}

void init_bucket (a)
  bucket_addr a;
{
  bucket_t b;

  b->first = Nil;
  b->self = a;
  write_bucket (b);
}

void find_bl (result, bucket, x, y)
  block_list_t result;
  bucket_t bucket;
  coord x, y;
{
  block_addr a;
  long depth;
  long i;

  a = bucket->first;
  for (depth = 0; depth < Deep_bucket; depth++){
    if (a == Nil){
      result->length = 0;
      return;
    }
    read_desc (result, a);
    if (Desc (result)->x == x && Desc (result)->y == y){
      read_blocks (result);
      return;
    }
    if (hash (Desc (result)->x, Desc (result)->y) != bucket->self){
      log ("find_bl: bad desc, bucket =", bucket->self);
    }
    a = Desc (result)->next_desc;
  }
  log ("find_bl: deep bucket", bucket->self);
  result->length = 0;
  return;
}

int find_last (rx, ry, bucket)
  coord *rx, *ry;
  bucket_t bucket;
{
  block_addr a;
  desc_block_t b;
  long depth;

  a = bucket->first;
  if (a == Nil) return 0;
  for (depth = 1; depth < Deep_bucket; depth++){
    read_block (b, a);
    if (hash (b->x, b->y) != bucket->self){
      log ("find_last: bad desc, bucket =", bucket->self);
    }
    a = b->next_desc;
    if (a == Nil){
      *rx = b->x;
      *ry = b->y;
      return 1;
    }
  }
  log ("find_last: deep bucket", bucket->self);
  return 0;
}

int find_first (rx, ry, bucket)
  coord *rx, *ry;
  bucket_t bucket;
{
  block_addr a;
  desc_block_t b;
  long depth;
  long n;

  a = bucket->first;
  if (a == Nil) return 0;
  read_block (b, a);
  if (hash (b->x, b->y) != bucket->self){
    log ("find_first: bad desc, bucket =", bucket->self);
    return 0;
  }
  *rx = b->x;
  *ry = b->y;
  return 1;
}

void remove_bl (result, d, bucket, x, y)
     block_list_t result;
     int *d;
     bucket_t bucket;
     coord x, y;
{
  block_addr a;
  desc_block_t prev;
  long depth;

  a = bucket->first;
  if (a == Nil){
    result->length = 0;
    Desc (result)->x = x;
    Desc (result)->y = y;
    *d = -1;
    return;
  }
  read_desc (result, a);
  if (Desc (result)->x == x && Desc (result)->y == y){
    bucket->first = Desc (result)->next_desc;
    write_bucket (bucket);
    read_blocks (result);
    *d = 0;
    return;
  }
  if (hash (Desc (result)->x, Desc (result)->y) != bucket->self){
    log ("remove_bl: bad desc 0, bucket =", bucket->self);
  }
  for (depth = 1; depth < Deep_bucket; depth++){
    if (Desc (result)->next_desc == Nil){
      result->length = 0;
      Desc (result)->x = x;
      Desc (result)->y = y;
      *d = -1;
      return;
    }
    *prev = *Desc(result);
    read_desc (result, prev->next_desc);
    if (Desc (result)->x == x && Desc (result)->y == y){
      read_blocks (result);
      prev->next_desc = Desc (result)->next_desc;
      write_block (prev);
      *d = depth;
      return;
    }
    if (hash (Desc (result)->x, Desc (result)->y) != bucket->self){
      log ("remove_bl: bad desc 1, bucket =", bucket->self);
    }
  }
  log ("remove_bl: deep bucket ", bucket->self);
  result->length = 0;
  Desc (result)->x = x;
  Desc (result)->y = y;
  *d = -1;
  return;
}

void insert_bl (bucket, bl, rank)
  bucket_t bucket;
  block_list_t bl;
  int rank;
{
  long i;
  desc_block_t prev;
  block_addr a;

  a = bucket->first;
  if (rank == 0 || a == Nil){
    Desc (bl)->next_desc = bucket->first;
    bucket->first = Desc (bl)->self;
    for (i = 0; i < bl->length; i++){
      write_block (&(bl->blocks [i]));
    }
    write_bucket (bucket);
  }else{
    while (rank > 0 && a != Nil){
      read_block (prev, a);
      a = prev->next_desc;
      --rank;
    }
    Desc (bl)->next_desc = prev->next_desc;
    prev->next_desc = Desc (bl)->self;
    for (i = 0; i < bl->length; i++){
      write_block (&(bl->blocks [i]));
    }
    write_block (prev);
  }
}

int check_bucket (b, bitmap)
     bucket_t b;
     char *bitmap;
{
  block_addr a;
  block_list_t bl;
  long n;
  long i;

  a = b->first;
  for (n = 0; n < Deep_bucket; n++){
    if (a == Nil) return n;
    read_desc (bl, a);
    check_blocks (bl);
    if (hash (Desc (bl)->x, Desc (bl)->y) != b->self){
      printf ("bogus bucket %x; desc block %x\n", b->self, Desc (bl)->self);
      return Deep_bucket;
    }
    bitmap [Desc (bl)->self] |= 2;
    a = Desc (bl)->next_desc;
  }
  printf ("deep bucket %x\n", b->self);
  return Deep_bucket;
}
