#include <fcntl.h>
#include <time.h>
#include <stdio.h>
#include "config.h"
#include "data.h"
#include "error.h"
#include "freelist.h"
#include "globals.h"
#include "hashtbl.h"

void create (mblocks, nbuckets, ro_x, ro_y)
  block_addr mblocks;
  bucket_addr nbuckets;
  coord ro_x, ro_y;
{
  unsigned long i;
  block_t b;

  misc_fd = open (Misc_file_name, O_RDWR | O_CREAT | O_EXCL, 0600);
  table_fd = open (Table_file_name, O_RDWR | O_CREAT | O_EXCL, 0600);
  data_fd = open (Data_file_name, O_RDWR | O_CREAT | O_EXCL, 0600);
  if (misc_fd < 0 || table_fd < 0 || data_fd < 0){
    perror ("open ");
    fatal (E_cant_open, "");
    exit (2);
  }

  if (mblocks <= 0) fatal (E_bad_param, "");
  if (nbuckets <= 0) fatal (E_bad_param, "");
  /* The number of buckets must be a prime number for the warp code to
     work properly. */
  for (i = 2; i < nbuckets; i++){
    if (nbuckets % i == 0){
      fatal (E_bad_param, "");
    }
  }

  I_max_blocks = mblocks;
  I_num_buckets = nbuckets;
  I_ro_x = ro_x;
  I_ro_y = ro_y;
  I_free_head = Nil;
  I_free_length = 0;
  I_epoch = time (0);
  I_read = 0;
  I_empty = 0;
  I_written = 0;
  I_erased = 0;
  I_added = 0;
  I_blasted = 0;
  I_cur_blocks = 0;
  init_info ();

  for (i = 0; i < nbuckets; i++){
    init_bucket (i);
  }
}

void dump ()
{
  brick_t brick;
  block_list_t bl;
  bucket_addr a;
  block_addr bla;
  bucket_t bk;
  long i;
  long depth;

  for (a = 0; a < I_num_buckets; a++){
    printf ("filt bucket number %lu\n", a);
    lock_bucket (bk, a);
    bla = bk->first;
    for (depth = 0; depth < Deep_bucket; depth++){
      if (bla == Nil) break;
      read_desc (bl, bla);
      read_blocks (bl);
      unpack_brick (brick, bl);
      printf ("writ %lu %lu\n", Desc (bl)->x, Desc (bl)->y);
      for (i = 0; i < 32; i++){
	printf ("%32.32s\n", brick [i]);
      }
      printf ("----\n");
      bla = Desc (bl)->next_desc;
    }
    unlock_bucket (bk);
  }
}
