/* Murof -- Public domain software written by Planar. */

#include <fcntl.h>
#include <termios.h>
#include <errno.h>
#include <stdio.h>
char *malloc ();
#include <murlib.h>
#include "command.h"
#include "config.h"
#include "error.h"
#include "globals.h"
#include "maintain.h"
#include "random.h"
#include "stats.h"

static void verify (fd)
  int fd;
{
  if (fd < 2){
    close (fd);
    errno = 0;
    fatal (E_std_closed, "");
  }
}

static void initialize (args)
     char *args;
{
  table_fd = open (Table_file_name, O_RDWR);
  if (table_fd < 0) fatal (E_cant_open, Table_file_name);
  data_fd = open (Data_file_name, O_RDWR);
  if (data_fd < 0) fatal (E_cant_open, Data_file_name);
  misc_fd = open (Misc_file_name, O_RDWR);
  if (misc_fd < 0) fatal (E_cant_open, Misc_file_name);

  read_info ();
  if (error) fatal (E_cant_read, Misc_file_name);
}

void verify_server ()
{
  /* murof.bin must be launched on the server because NFS locking is bogus. */
  char buf [Host_name_len];

  buf [Host_name_len - 1] = '\0';
  gethostname (buf, Host_name_len - 1);
  if (strcmp (buf, Server_host_name)){
    errno = 0;
    fatal (E_server_only, Server_host_name);
  }
}

int main (argc, argv)
  int argc;
  char *argv [];
{
  c = init_mur_connection (0, 1, 0);
  if (c == NULL){
    perror ("cannot initialize connection");
    exit (2);
  }

  log_fd = open (Log_file_name, O_RDWR | O_CREAT | O_APPEND, 0600);
  if (log_fd < 0) fatal (E_cant_open, Log_file_name);
  verify (log_fd);
  random_init ();

  verify_server ();
  if (argc > 1){
    if (!strcmp ("-C", argv [1])){
      if (argc != 6){
	errno = 0;
	fatal (E_usage, "");
      }
      create (atoi (argv [2]), atoi (argv [3]), atoi (argv [4]),
	      atoi (argv [5]));
      unrestricted = 1;
      /* Fall through to read-execute loop. */
    }else if (!strcmp ("-D", argv [1])){
      if (argc != 2){
	errno = 0;
	fatal (E_usage, "");
      }
      initialize ("");
      dump ();
      exit (0);
    }else if (!strcmp ("-A", argv [1])){
      if (argc != 2){
	errno = 0;
	fatal (E_usage, "");
      }
      unrestricted = 1;
      initialize ("");
      /* Fall through to read-execute loop. */
    }else{
      errno = 0;
      fatal (E_usage, "");
    }
  }else{
    /* Normal use. */
    if (open (Nomur_file_name, O_RDONLY) >= 0){
      errno = 0;
      fatal (E_nomur, "");
    }
    initialize ("");
    com_warp ();
    if (error) fatal (E_cant_initialize, error_msg ());
  }

  /* Read-execute loop. */
  { message m;
    int is_empty;

    m.type = rep_helo;
    m.x = Version_x;
    m.y = Version_y;

    while (1){
      if (mur_write (c, &m) == -1){
	perror ("cannot write to stdout");
	exit (2);
      }
      if (m.type == rep_bye) break;
      if (mur_read (c, &m) == -1){
	perror ("cannot read from stdin");
	exit (2);
      }
      error = 0;
      switch (m.type){
      case req_read:
        is_empty = com_read (m.data, m.x, m.y);
        if (error) goto nonfatal_error;
        if (is_empty){
	  m.type = rep_empt;
	}else{
	  m.type = rep_brik;
	}
        break;
      case req_writ:
        com_write (m.x, m.y, m.data);
        if (error) goto nonfatal_error;
        m.type = rep_wrtn;
        break;
      case req_eras:
        fill_brick (m.data);
        com_write (m.x, m.y, m.data);
        if (error) goto nonfatal_error;
        m.type = rep_wrtn;
        break;
      case req_warp:
        com_warp ();
        if (error) goto nonfatal_error;
        m.type = rep_wrpd;
        break;
      case req_stat:
	sprint_stats (m.data);
	if (error) goto nonfatal_error;
	m.type = rep_stts;
	break;
      case m_eof:
      case req_exit:
        m.type = rep_bye;
        break;
      case req_filt:
	strcpy (m.data, "Filter function not implemented.");
	m.type = rep_fltr;
	break;
      case m_bug:
	m.type = rep_erro;
	break;
      default:
	strcpy (m.data, "do not send replies to the server");
	m.x = E_syntax_type;
        m.type = rep_erro;
	break;
      }
      continue;
    nonfatal_error:
      strcpy (m.data, error_msg (error));
      m.x = error;
      m.type = rep_erro;
    }
  }
  merge_stats ();
  exit (0);
}
